@@ -1,4 +1,4 @@
-# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
+# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
 # SPDX-License-Identifier: Apache-2.0
 
 """
@@ -16,7 +16,8 @@ For a higher-level, more declarative API, see :mod:`dazl.query`.
 """
 
 from . import config
-from .api import simple_client, Network, PartyClient, AIOPartyClient, SimplePartyClient
+from .api import async_network, simple_client, Network, PartyClient, AIOPartyClient, \
+    SimplePartyClient
 from .bots import Bot, BotCollection, BotEntry
 from .runner import run
 from ._base_model import ExitCode, LedgerRun, CREATE_IF_MISSING, NONE_IF_MISSING, \
