@@ -1,49 +1,51 @@
-# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
+# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
 # SPDX-License-Identifier: Apache-2.0
 
 import logging
-from asyncio import gather, ensure_future, sleep
+from asyncio import gather, ensure_future, sleep, wait_for
 from collections import defaultdict
-from dataclasses import fields
-from datetime import timedelta, datetime
+from dataclasses import asdict, fields
+from datetime import timedelta
 from threading import RLock, Thread, current_thread
 from typing import Any, Callable, Dict, Optional, TypeVar, Collection, Awaitable, Set, Tuple, \
     Union, overload
+
 try:
     from typing import Literal
 except ImportError:
     from typing_extensions import Literal
 
-from dataclasses import asdict
-
 from .. import LOG
 from ._base_model import IfMissingPartyBehavior, CREATE_IF_MISSING, NONE_IF_MISSING, \
     EXCEPTION_IF_MISSING
 from ._party_client_impl import _PartyClientImpl
 from .bots import Bot, BotCollection
 from .config import AnonymousNetworkConfig, NetworkConfig, URLConfig
+from ..damlast.lookup import MultiPackageLookup
+from ..damlast.pkgfile import get_dar_package_ids
 from ..metrics import MetricEvents
 from ..model.core import Party, DazlPartyMissingError
 from ..model.ledger import LedgerMetadata
 from ..model.network import connection_settings
 from ..model.reading import InitEvent, ReadyEvent, BaseEvent
+from ..prim.datetime import TimeDeltaLike, to_timedelta
 from ..protocols import LedgerNetwork
 from ..protocols.autodetect import AutodetectLedgerNetwork
 from ..scheduler import Invoker, RunLevel
-from ..util.dar import get_dar_package_ids
-from ..util.prim_types import to_timedelta, TimeDeltaConvertible
 
 T = TypeVar('T')
 
 
 class _NetworkImpl:
 
-    __slots__ = ('_lock', '_main_thread', 'invoker', '_party_clients', '_global_impls',
-                 '_party_impls', 'bots', '_config', '_pool', '_pool_init',
+    __slots__ = ('lookup', 'invoker', '_lock', '_main_thread', '_party_clients',
+                 '_global_impls', '_party_impls', 'bots', '_config', '_pool', '_pool_init',
                  '_cached_metadata', '_metrics')
 
     def __init__(self, metrics: 'Optional[MetricEvents]' = None):
+        self.lookup = MultiPackageLookup()
         self.invoker = Invoker()
+
         self._lock = RLock()
         self._main_thread = None  # type: Optional[Thread]
         self._pool = None  # type: Optional[LedgerNetwork]
@@ -87,6 +89,43 @@ class _NetworkImpl:
     def resolved_anonymous_config(self) -> 'AnonymousNetworkConfig':
         return AnonymousNetworkConfig.parse_kwargs(**self._config)
 
+    def freeze(self):
+        """
+        Freeze configuration, and assume the current run loop can be used to schedule dazl
+        coroutines. Once this method is called, ``aio_run()`` must be called instead of ``start()``
+        in order to run dazl.
+        """
+        with self._lock:
+            self.invoker.set_context_as_current()
+            config = self.resolved_config()
+            LOG.debug("Resolved config: %s", config)
+
+        # From this point on, we're assuming we're on an asyncio event loop so locking is no longer
+        # required
+        if self._pool is None:
+            from ..protocols import LedgerConnectionOptions
+
+            # If the connect timeout is non-positive, assume the user intended for there to not be
+            # a timeout at all
+            connect_timeout = to_timedelta(config.connect_timeout)
+            if connect_timeout <= timedelta(0):
+                connect_timeout = None
+
+            package_lookup_timeout = to_timedelta(config.package_lookup_timeout)
+            if package_lookup_timeout <= timedelta(0):
+                package_lookup_timeout = None
+
+            options = LedgerConnectionOptions(
+                lookup=self.lookup,
+                connect_timeout=connect_timeout,
+                package_lookup_timeout=package_lookup_timeout,
+                eager_package_fetch=config.eager_package_fetch)
+
+            self._pool = pool = AutodetectLedgerNetwork(self.invoker, options)
+            self._pool_init.set_result(pool)
+
+        return config
+
     async def aio_run(self, *coroutines) -> None:
         """
         Coroutine where all network activity is scheduled from.
@@ -96,11 +135,8 @@ class _NetworkImpl:
             of the additional coroutines are also finished.
         """
         from ..metrics.instrumenters import AioLoopPerfMonitor
-        from ..protocols import LedgerConnectionOptions
 
-        with self._lock:
-            self.invoker.set_context_as_current()
-            config = self.resolved_config()
+        config = self.freeze()
 
         site = None
         with AioLoopPerfMonitor(self._metrics.loop_responsiveness):
@@ -118,19 +154,8 @@ class _NetworkImpl:
                 LOG.info('No server_port configuration was specified, so metrics and other stats '
                          'will not be served.')
 
-            # If the connect timeout is non-positive, assume the user intended for there to not be
-            # a timeout at all
-            connect_timeout = to_timedelta(config.connect_timeout)
-            if connect_timeout <= timedelta(0):
-                connect_timeout = None
-
-            options = LedgerConnectionOptions(connect_timeout=connect_timeout)
-
-            self._pool = pool = AutodetectLedgerNetwork(self.invoker, options)
-            self._pool_init.set_result(pool)
-
             try:
-                runner = _NetworkRunner(pool, self, coroutines)
+                runner = _NetworkRunner(self._pool, self, coroutines)
                 await runner.run()
             finally:
                 await self._pool.close()
@@ -291,7 +316,7 @@ class _NetworkImpl:
                     return bot
         return None
 
-    def simple_metadata(self, timeout: 'TimeDeltaConvertible') -> LedgerMetadata:
+    def simple_metadata(self, timeout: 'TimeDeltaLike') -> 'LedgerMetadata':
         """
         Retrieve metadata about the ledger.
 
@@ -304,48 +329,34 @@ class _NetworkImpl:
         with self._lock:
             return self.invoker.run_in_loop(self._pool.ledger, timeout=timeout)
 
-    async def aio_metadata(self) -> LedgerMetadata:
+    async def aio_metadata(self) -> 'LedgerMetadata':
         """
         Coroutine version of :meth:`metadata`.
         """
         pool = await self._pool_init
         return await pool.ledger()
 
-    async def upload_package(self, contents: bytes, timeout: 'TimeDeltaConvertible') -> None:
+    async def upload_package(self, contents: bytes, timeout: 'TimeDeltaLike') -> None:
         """
         Ensure packages specified by the given byte array are loaded on the remote server. This
-        method only returns once packages are reported by the package services.
-
-        If an admin URL is specified, dazl will attempt to upload packages to that URL, and wait
-        for the package to be reported by the package service.
+        method only returns once packages are reported by the Package Service.
 
         :param contents: Contents to attempt to upload.
         :param timeout: Length of time before giving up.
         """
         pool = await self._pool_init
-        package_ids = get_dar_package_ids(contents)
+        await self.connect_anonymous()
+        package_ids = await self.invoker.run_in_executor(lambda: get_dar_package_ids(contents))
         await pool.upload_package(contents)
         await self.ensure_package_ids(package_ids, timeout)
 
     async def ensure_package_ids(
-            self, package_ids: 'Collection[str]', timeout: 'TimeDeltaConvertible'):
-        from asyncio import wait_for, TimeoutError
-        timeout = to_timedelta(timeout)
-        expire_time = datetime.max #datetime.utcnow() + timeout
-        metadata = await wait_for(self.aio_metadata(), timeout.total_seconds())
-        package_id_set = set(package_ids)
-        while datetime.utcnow() < expire_time:
-            if package_id_set.issubset(set(metadata.store.package_ids())):
-                return
-        raise TimeoutError()
+            self, package_ids: 'Collection[str]', timeout: 'TimeDeltaLike'):
+        await wait_for(self.__ensure_package_ids(package_ids), to_timedelta(timeout).total_seconds())
 
-    async def get_time(self) -> datetime:
-        metadata = await self._pool.ledger()
-        await self._pool.sync_time()
-        return metadata.time_model.get_time()
-
-    def set_time(self, new_datetime: datetime) -> Awaitable[None]:
-        return self._pool.set_time(new_datetime)
+    async def __ensure_package_ids(self, package_ids: 'Collection[str]'):
+        metadata = await self.aio_metadata()
+        await gather(*(metadata.package_loader.load(pkg_ref) for pkg_ref in package_ids))
 
     # region Event Handler Management
 
@@ -372,7 +383,8 @@ class _NetworkImpl:
             verify_ssl=config.verify_ssl,
             ca_file=config.ca_file,
             cert_file=config.cert_file,
-            cert_key_file=config.cert_key_file)
+            cert_key_file=config.cert_key_file,
+            enable_http_proxy=config.enable_http_proxy)
         await self._pool.connect_anonymous(settings, url_prefix)
 
 
@@ -491,30 +503,30 @@ class _NetworkRunner:
 
         # Raise the 'init' event.
         init_futs = []
-        current_time = metadata.time_model.get_time()
         if first_offset is None:
-            evt = InitEvent(None, None, current_time, metadata.ledger_id, metadata.store)
+            evt = InitEvent(
+                None, None, None, metadata.ledger_id, self._network_impl.lookup, metadata._store)
             init_futs.append(ensure_future(self._network_impl.emit_event(evt)))
         for party_impl in party_impls:
-            init_futs.append(ensure_future(party_impl.initialize(current_time, metadata)))
+            init_futs.append(ensure_future(party_impl.initialize(None, metadata)))
 
         # TODO: Consider what should join on init_futs; we can't join here because otherwise this
         #  blocks the readers from reading transactions that would cause these events to fully
         #  resolve
 
         from ._reader_sync import read_initial_acs
-        use_acs_service = self._config.use_acs_service
-        offset = await read_initial_acs(party_impls, use_acs_service)
+        offset = await read_initial_acs(party_impls)
 
         LOG.debug('Preparing to raise the "ready" event...')
         # Raise the 'ready' event.
         ready_futs = []
-        current_time = metadata.time_model.get_time()
         if first_offset is None:
-            evt = ReadyEvent(None, None, current_time, metadata.ledger_id, metadata.store, offset)
+            evt = ReadyEvent(
+                None, None, None, metadata.ledger_id, self._network_impl.lookup, metadata._store,
+                offset)
             ready_futs.append(ensure_future(self._network_impl.emit_event(evt)))
         for party_impl in party_impls:
-            ready_futs.append(ensure_future(party_impl.emit_ready(metadata, current_time, offset)))
+            ready_futs.append(ensure_future(party_impl.emit_ready(metadata, None, offset)))
         for party_impl in party_impls:
             party_impl.ready().set_result(None)
 
