@@ -1,11 +1,11 @@
-# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
+# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
 # SPDX-License-Identifier: Apache-2.0
 
 from asyncio import ensure_future, gather, get_event_loop, wait, Future
 from concurrent.futures import ALL_COMPLETED
 
 from dataclasses import replace, dataclass, field, fields, asdict
-from datetime import datetime, timedelta
+from datetime import datetime
 from io import StringIO
 from typing import Any, Awaitable, Collection, List, Optional, Sequence, Set, Tuple, \
     TypeVar, TYPE_CHECKING
@@ -17,18 +17,19 @@ from .bots import Bot, BotCollection, BotCallback, BotFilter
 from .config import NetworkConfig, PartyConfig
 from .state import ActiveContractSet
 from ._writer_verify import ValidateSerializer
+from ..damlast.daml_lf_1 import TypeConName
 from ..model.core import ContractMatch, ContractsState, ContractId, \
-    CommandTimeoutError, ContractContextualData, ContractContextualDataCollection, Party
+    ContractContextualData, ContractContextualDataCollection, Party
 from ..model.ledger import LedgerMetadata
-from ..model.network import connection_settings
+from ..model.network import connection_settings, OAuthSettings
 from ..model.reading import BaseEvent, TransactionStartEvent, TransactionEndEvent, OffsetEvent, \
     TransactionFilter, ContractCreateEvent, ContractArchiveEvent, \
-    InitEvent, ReadyEvent, ActiveContractSetEvent, PackagesAddedEvent
+    InitEvent, ReadyEvent, ActiveContractSetEvent, PackagesAddedEvent, ContractFilter
 from ..model.writing import CommandBuilder, CommandDefaults, CommandPayload, EventHandlerResponse
+from ..prim import TimeDeltaLike, to_timedelta
 from ..protocols import LedgerNetwork, LedgerClient
-from ..util.asyncio_util import ServiceQueue, await_then, completed, named_gather
+from ..util.asyncio_util import ServiceQueue, completed, named_gather
 from ..util.prim_natural import n_things
-from ..util.prim_types import TimeDeltaConvertible, to_timedelta
 from ..util.typing import safe_cast
 
 if TYPE_CHECKING:
@@ -54,20 +55,38 @@ class _PartyClientImpl:
         self._ready_fut = self.invoker.create_future()
         self._known_packages = set()  # type: Set[str]
 
-        self._acs = ActiveContractSet(self.invoker)
+        self._acs = ActiveContractSet(self.invoker, self.parent.lookup)
         self.bots = BotCollection(party)
         self._reader = _PartyClientReaderState()
         self._writer = _PartyClientWriterState()
 
     def connect_in(self, pool: 'LedgerNetwork') -> 'Future':
         self._config = config = self.resolved_config()
+
+        oauth=None
+        if config.oauth_token or config.oauth_client_id:
+            oauth=OAuthSettings(
+                token=config.oauth_token,
+                token_uri=config.oauth_token_uri,
+                refresh_token=config.oauth_refresh_token,
+                client_id=config.oauth_client_id,
+                client_secret=config.oauth_client_secret,
+                id_token=config.oauth_id_token,
+                redirect_uri=config.oauth_redirect_uri,
+                auth_url=config.oauth_auth_url,
+                auth_audience=config.oauth_audience,
+                auth_ca_file=config.oauth_ca_file
+            )
+
         settings, url_prefix = connection_settings(
             config.url,
             self.party,
             verify_ssl=config.verify_ssl,
             ca_file=config.ca_file,
             cert_file=config.cert_file,
-            cert_key_file=config.cert_key_file)
+            cert_key_file=config.cert_key_file,
+            oauth=oauth
+        )
 
         self._pool = pool
         self._client_fut = ensure_future(pool.connect(self.party, settings, url_prefix))
@@ -96,19 +115,13 @@ class _PartyClientImpl:
         :param metadata:
             Information about the connected ledger.
         """
-        self._acs.metadata_future.set_result(metadata.store)
-        evt = InitEvent(self, self.party, current_time, metadata.ledger_id, metadata.store)
+        evt = InitEvent(
+            self, self.party, current_time, metadata.ledger_id, self.parent.lookup, metadata._store)
         return self.emit_event(evt)
 
     def ready(self) -> Awaitable[None]:
         return self._ready_fut
 
-    def get_time(self) -> Awaitable[datetime]:
-        return await_then(self._pool.ledger(), lambda metadata: metadata.time_model.get_time())
-
-    def set_time(self, new_datetime) -> Awaitable[None]:
-        return await_then(self._pool_fut, lambda pool: pool.set_time(new_datetime))
-
     # region Event Handler Management
 
     # noinspection PyShadowingBuiltins
@@ -147,11 +160,13 @@ class _PartyClientImpl:
             Nothing. This method "returns" when all events have been raised and their follow-ups
             have been processed.
         """
-        ready_event = ReadyEvent(self, self.party, time, metadata.ledger_id, metadata.store, offset)
-        self._known_packages.update(metadata.store.package_ids())
+        ready_event = ReadyEvent(
+            self, self.party, time, metadata.ledger_id, self.parent.lookup, metadata._store, offset)
+        self._known_packages.update(self.parent.lookup.package_ids())
         await self.emit_event(ready_event)
 
-        pkg_event = PackagesAddedEvent(self, self.party, time, metadata.ledger_id, metadata.store, True)
+        pkg_event = PackagesAddedEvent(
+            self, self.party, time, metadata.ledger_id, self.parent.lookup, metadata._store, True)
         await self.emit_event(pkg_event)
 
     # endregion
@@ -184,7 +199,7 @@ class _PartyClientImpl:
     # region Read Path
 
     async def read_acs(self, until_offset: 'Optional[str]', raise_events: bool) \
-            -> 'Tuple[str, Future]':
+            -> 'Tuple[Optional[str], Future]':
         """
         Initial bootstrap of events from the read side. Only one instance of this coroutine
         should be active at a time per client. An initial block of events is read using the
@@ -201,12 +216,28 @@ class _PartyClientImpl:
         :return:
             The ledger offset that the reader ended at and a Future that is resolved when all event
             handlers' follow-ups have either successfully or unsuccessfully completed.
+
+            If the PartyClient is _not_ configured to use the ActiveContractService, this function
+            immediately returns `None`, and a resolved future.
         """
-        LOG.info('Calling read_acs(%r, %r)', until_offset, raise_events)
+        if not self._config.use_acs_service:
+            LOG.info('read_acs(%r, %r) aborted for party %r because it was configured to not use '
+                     'the ACS service', until_offset, raise_events, self.party)
+            from asyncio import get_event_loop
+            fut = get_event_loop().create_future()
+            fut.set_result(None)
+            return None, fut
+
+        LOG.info('Calling read_acs(%r, %r) for party: %r', until_offset, raise_events, self.party)
         client = await self._client_fut
-        acs = await client.active_contracts()
+
+        contract_filter = ContractFilter(
+            templates=await self._template_filter(), party_groups=self._config.party_groups)
+
+        acs = await client.active_contracts(contract_filter)
         for acs_evt in acs:
             await self._process_transaction_stream_event(acs_evt, False)
+        LOG.info('read_acs() finished at offset %r for party: %r', self._reader.offset, self.party)
 
         return await self.read_transactions(until_offset, raise_events)
 
@@ -246,21 +277,20 @@ class _PartyClientImpl:
             # TODO: Find a more appropriate place to raise these events (or change the name of this
             #  method to make it clearer that it has a bigger mandate than simply transaction
             #  events)
-            all_packages = set(metadata.store.package_ids())
+            all_packages = set(self.parent.lookup.package_ids())
             new_package_ids = all_packages - self._known_packages
             if new_package_ids:
                 pkg_event = PackagesAddedEvent(
-                    self, self.party, None, metadata.ledger_id, metadata.store, False)
+                    self, self.party, None, metadata.ledger_id, self.parent.lookup, metadata._store,
+                    False)
                 self._known_packages.update(all_packages)
                 await self.emit_event(pkg_event)
 
             # prepare a call to /events
             transaction_filter = TransactionFilter(
-                ledger_id=metadata.ledger_id,
                 current_offset=self._reader.offset,
                 destination_offset=until_offset,
-                templates=None,
-                max_blocks=None,
+                templates=await self._template_filter(),
                 party_groups=self._config.party_groups)
 
             transaction_events = await client.events(transaction_filter)
@@ -288,6 +318,19 @@ class _PartyClientImpl:
         client = await self._client_fut
         return await client.events_end()
 
+    async def _template_filter(self) -> 'Optional[Collection[TypeConName]]':
+        if self._config.package_ids is not None:
+            # first, ensure the packages we require are loaded
+            metadata = await self._pool.ledger()
+            await gather(*[metadata.package_loader.load(ref) for ref in self._config.package_ids])
+
+            # now build a filter of the template names for all of the packages that we found
+            names = [name for ref in self._config.package_ids
+                     for name in self.parent.lookup.template_names(f"{ref}:*")]
+            return names
+        else:
+            return None
+
     def _process_transaction_stream_event(self, event: Any, raise_events: bool) -> Future:
         """
 
@@ -307,7 +350,7 @@ class _PartyClientImpl:
             # Notify anything waiting on commands to complete that some of them will have
             # completed as a consequence of this transaction.
             for cmd in self._writer.inflight_commands:
-                cmd.notify_read_done(event.command_id, event.time)
+                cmd.notify_read_done(event.command_id)
 
             LOG.debug('evt recv: party %s, BIM %r (%s events)',
                       self.party, event.command_id[0:7], len(event.contract_events))
@@ -335,7 +378,7 @@ class _PartyClientImpl:
             commands: EventHandlerResponse,
             ignore_errors: bool = False,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> Awaitable[None]:
         """
         Submit a command or list of commands.
@@ -347,8 +390,10 @@ class _PartyClientImpl:
             then a failure to send this command does not necessarily end the client.
         :param workflow_id:
             The workflow ID to use to tag submitted commands.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         :return:
             An ``asyncio.Future`` that is resolved right before the corresponding side effects have
             hit the event stream for this party. Synchronous errors are reported back immediately
@@ -357,7 +402,10 @@ class _PartyClientImpl:
         if workflow_id is None:
             workflow_id = uuid.uuid4().hex
         cb = CommandBuilder.coerce(commands, atomic_default=True)
-        cb.defaults(workflow_id=workflow_id, ttl=to_timedelta(ttl) if ttl is not None else None)
+        cb.defaults(
+            workflow_id=workflow_id,
+            deduplication_time=(to_timedelta(deduplication_time)
+                                if deduplication_time is not None else None))
 
         p = _PendingCommand(cb)
         p.future.add_done_callback(lambda _: self._process_command_finished(p, ignore_errors))
@@ -367,16 +415,20 @@ class _PartyClientImpl:
 
     def _process_command_finished(self, pending_command, ignore_errors):
         try:
-            if pending_command.future.exception() is None:
-                LOG.debug('Command finished: %s', pending_command)
             self._writer.inflight_commands.remove(pending_command)
         except ValueError:
             LOG.warning('Tried to remove %s even though it was already removed.', pending_command)
 
-        if pending_command.future.exception() is not None:
-            # TODO: more with this; maybe let the user respond to this
-            LOG.exception('A command submission failed!',
-                          exc_info=pending_command.future.exception())
+        if pending_command.future.cancelled():
+            LOG.debug("Command cancelled: %s", pending_command)
+        else:
+            ex = pending_command.future.exception()
+            if ex is not None:
+                # TODO: more with this; maybe let the user respond to this
+                LOG.exception('A command submission failed!',
+                              exc_info=pending_command.future.exception())
+            else:
+                LOG.debug('Command finished: %s', pending_command)
 
     async def main_writer(self):
         """
@@ -387,7 +439,7 @@ class _PartyClientImpl:
 
         client = self._client_fut.result()  # type: LedgerClient
         metadata = ledger_fut.result()  # type: LedgerMetadata
-        validator = ValidateSerializer(metadata.store)
+        validator = ValidateSerializer(self.parent.lookup)
 
         self._writer.pending_commands.start()
 
@@ -397,9 +449,14 @@ class _PartyClientImpl:
         # ServiceQueue is stopped.
         async for p in self._writer.pending_commands:
             LOG.debug('Sending a command: %s', p)
-            ledger_effective_time = metadata.time_model.get_time()
             command_payloads = []  # type: List[Tuple[_PendingCommand, Sequence[CommandPayload]]]
 
+            if p.future.done():
+                # PendingCommand instances that are already marked as done have either been marked
+                # as a failure or cancelled by the caller. Do NOT send the corresponding Ledger API
+                # command because the PendingCommand() has effectively been aborted.
+                continue
+
             self._writer.inflight_commands.append(p)
             try:
                 defaults = CommandDefaults(
@@ -407,18 +464,18 @@ class _PartyClientImpl:
                     default_ledger_id=metadata.ledger_id,
                     default_workflow_id=None,
                     default_application_id=self._config.application_name,
-                    default_command_id=None,
-                    default_ttl=timedelta(seconds=30))
-                cps = p.command.build(defaults, ledger_effective_time)
+                    default_command_id=None)
+                cps = p.command.build(defaults)
                 if cps:
-                    commands = [replace(cp, commands=validator.serialize_commands(cp.commands))
-                                for cp in cps]  # type: Sequence[CommandPayload]
+                    commands = await metadata.package_loader.do_with_retry(
+                        lambda: [replace(cp, commands=validator.serialize_commands(cp.commands)) for cp in cps])
                     command_payloads.append((p, commands))
                     await submit_command_async(client, p, commands)
                 else:
                     # This is a "null command"; don't even bother sending to the server. Immediately
                     # resolve the future successfully and discard
-                    p.future.set_result(None)
+                    if not p.future.done():
+                        p.future.set_result(None)
             except Exception as ex:
                 LOG.exception("Tried to send a command and failed!")
                 p.notify_read_fail(ex)
@@ -465,21 +522,17 @@ class _PendingCommand:
     Track the status of a set of commands in-flight.
     """
 
-    __slots__ = ('command', 'command_ids', 'max_record_time', 'future')
+    __slots__ = ('command', 'command_ids', 'future')
 
     def __init__(self, command: CommandBuilder):
         self.command = safe_cast(CommandBuilder, command)
         self.command_ids = None  # type: Optional[Set[str]]
-        self.max_record_time = None  # type: Optional[datetime]
         self.future = get_event_loop().create_future()
 
-    def notify_write(self, command_ids: Collection[str], max_record_time: datetime):
-        if self.max_record_time is not None:
-            raise Exception('cannot send an already in-progress command')
+    def notify_write(self, command_ids: Collection[str]):
         self.command_ids = set(command_ids)
-        self.max_record_time = max_record_time
 
-    def notify_read_done(self, command_id: str, ledger_time: Optional[datetime]):
+    def notify_read_done(self, command_id: str):
         """
         Trigger an appropriate response given the receipt of a transaction.
         """
@@ -490,15 +543,20 @@ class _PendingCommand:
             self.command_ids.discard(command_id)
             if not self.command_ids:
                 # the command is finished
-                self.future.set_result(None)
+                if not self.future.done():
+                    self.future.set_result(None)
                 return
 
-        if self.max_record_time is not None and ledger_time is not None and \
-                self.max_record_time < ledger_time:
-            self.future.set_exception(CommandTimeoutError())
-
     def notify_read_fail(self, ex: Exception):
-        self.future.set_exception(ex)
+        if self.future.done():
+            # The user may have cancelled the command or otherwise terminated the Future; trying
+            # to set an exception on a done Future will itself throw an exception, so the best
+            # we can do is log
+            LOG.exception(
+                'An exception was received and reported to a Future that has already been aborted! '
+                'This cannot be reported in a normal way.', ex)
+        else:
+            self.future.set_exception(ex)
 
     def __eq__(self, other):
         return self is other
@@ -513,10 +571,6 @@ class _PendingCommand:
             if self.command_ids is not None:
                 buf.write(', command_ids=')
                 buf.write(format(self.command_ids))
-            if self.max_record_time is not None:
-                buf.write(', max_record_time="')
-                buf.write(format(self.max_record_time.isoformat()))
-                buf.write('"')
             buf.write(', future=')
             buf.write(format(self.future))
             buf.write('>')
@@ -540,7 +594,6 @@ def submit_command_async(
     """
     coros = []
     command_ids = []
-    maximum_record_time = commands[0].maximum_record_time
 
     for payload in commands:
         coro = None
@@ -558,7 +611,7 @@ def submit_command_async(
         if coro is not None:
             coros.append(coro)
 
-    p.notify_write(command_ids, maximum_record_time)
+    p.notify_write(command_ids)
     if len(coros) == 0:
         return completed(None)
     elif len(coros) == 1:
