@@ -1,4 +1,4 @@
-# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
+# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
 # SPDX-License-Identifier: Apache-2.0
 
 """
@@ -17,40 +17,38 @@ specific party.
 # for a more concise representation of the various flavors of the API. The unit test
 # ``test_api_consistency.py`` verifies that these implementations are generally in sync with each
 # other the way that the documentation says they are.
-from asyncio import get_event_loop
-from contextlib import contextmanager, ExitStack
-from datetime import datetime
+from asyncio import get_event_loop, ensure_future
+from contextlib import contextmanager
 from functools import wraps
 from logging import INFO
 from pathlib import Path
 from uuid import uuid4
-from threading import current_thread, main_thread
 from typing import Any, Awaitable, BinaryIO, Collection, ContextManager, List, Optional, \
     Tuple, Union
-from urllib.parse import urlparse
 
 from .. import LOG
 from .bots import Bot, BotCollection
 from .config import AnonymousNetworkConfig, NetworkConfig, PartyConfig
 from ._base_model import IfMissingPartyBehavior, CREATE_IF_MISSING
-from ..damlsdk.sandbox import sandbox
+from ..damlast import get_dar_package_ids
+from ..damlast.protocols import SymbolLookup
 from ..metrics import MetricEvents
-from ..model.core import ContractId, ContractData, ContractsState, ContractMatch, \
-    ContractContextualData, ContractContextualDataCollection, Party
+from ..model.core import ContractsState, ContractMatch, \
+    ContractContextualData, ContractContextualDataCollection, Dar
 from ..model.ledger import LedgerMetadata
 from ..model.reading import InitEvent, ReadyEvent, ContractCreateEvent, ContractExercisedEvent, \
     ContractArchiveEvent, TransactionStartEvent, TransactionEndEvent, PackagesAddedEvent, EventKey
 from ..model.types import TemplateNameLike
 from ..model.writing import EventHandlerResponse
+from ..prim import ContractId, ContractData, Party, TimeDeltaLike, to_party
 from ..scheduler import RunLevel, validate_install_signal_handlers
 from ..util.asyncio_util import await_then
 from ..util.io import get_bytes
-from ..util.prim_types import TimeDeltaConvertible
 from ._events import EventHandler, AEventHandler, EventHandlerDecorator, AEventHandlerDecorator, \
     fluentize
 from ._network_client_impl import _NetworkImpl
 from ._party_client_impl import _PartyClientImpl
-
+from ..util.tools import as_list
 
 DEFAULT_TIMEOUT_SECONDS = 30
 
@@ -82,8 +80,6 @@ def simple_client(url: 'Optional[str]' = None, party: 'Union[None, str, Party]'
     import os
     if url is None:
         url = os.getenv('DAML_LEDGER_URL')
-    if party is None:
-        party = os.getenv('DAML_LEDGER_PARTY') or uuid4().hex
     if not url:
         raise ValueError('url must be specified, or the DAML_LEDGER_URL environment variable '
                          'must be set')
@@ -91,30 +87,56 @@ def simple_client(url: 'Optional[str]' = None, party: 'Union[None, str, Party]'
         raise ValueError('party must be specified, or the DAML_LEDGER_PARTY environment variable '
                          'must be set')
 
-    with ExitStack() as context_manager:
-        LOG.info('Starting a simple_client with to %s with party %r...', url, party)
-        parsed_url = urlparse(url)
-        if parsed_url.scheme is not None and parsed_url.scheme == 'sandbox':
-            # start a local in-memory sandbox first
-            daml_path = Path(os.getenv('DAML_LEDGER_DAR_PATH', 'target'))
+    LOG.info('Starting a simple_client with to %s with party %r...', url, party)
 
-            daml_artifacts = []  # type: List[Path]
-            daml_artifacts.extend(daml_path.glob('**/*.dar'))
-            daml_artifacts.extend(daml_path.glob('**/*.dalf'))
+    network = Network()
+    network.set_config(url=url)
+    client = network.simple_party(party) if party else network.simple_new_party()
 
-            sandbox_proc = sandbox(daml_path=daml_artifacts)
-            url = context_manager.enter_context(sandbox_proc).url
+    network.start_in_background()
 
-        network = Network()
-        network.set_config(url=url)
-        client = network.simple_party(party)
+    yield client
 
-        network.start_in_background()
+    network.shutdown()
+    network.join()
 
-        yield client
 
-        network.shutdown()
-        network.join()
+# This class is intended to be used as a function.
+# noinspection PyPep8Naming
+class async_network:
+    """
+    Create a :class:`Network` and ensure that it has the given set of DARs loaded.
+    """
+    def __init__(
+            self,
+            url: 'Optional[str]' = None,
+            dars: 'Optional[Union[Dar, Collection[Dar]]]' = None):
+        LOG.debug('async_network.__init__')
+        self.network = Network()
+        if url:
+            self.network.set_config(url=url)
+        self.dars = as_list(dars)  # type: List[Dar]
+
+        LOG.debug("Analyzing package_id config...")
+        self.package_ids = {pkg_id for dar in self.dars for pkg_id in get_dar_package_ids(dar)} \
+            if self.dars else None
+        LOG.debug("Package id analysis done.")
+
+        if self.package_ids:
+            self.network.set_config(package_ids=self.package_ids)
+
+    async def __aenter__(self):
+        LOG.debug('async_network.__aenter__')
+
+        for dar in self.dars:
+            await self.network.aio_global().ensure_dar(dar)
+        return self.network
+
+    async def __aexit__(self, exc_type, exc_val, exc_tb):
+        LOG.debug('async_network.__aexit__')
+        fut = self.network.shutdown()
+        if fut is not None:
+            await fut
 
 
 class Network:
@@ -125,6 +147,7 @@ class Network:
 
     def __init__(self, metrics: 'Optional[MetricEvents]' = None):
         self._impl = _NetworkImpl(metrics)
+        self._main_fut = None
 
     def set_config(
             self,
@@ -134,6 +157,14 @@ class Network:
             **kwargs):
         self._impl.set_config(*config, url=url, admin_url=admin_url, **kwargs)
 
+    @property
+    def lookup(self) -> 'SymbolLookup':
+        """
+        Return a :class:`SymbolLookup` that provides type and package information for known
+        packages.
+        """
+        return self._impl.lookup
+
     def resolved_config(self) -> 'NetworkConfig':
         """
         Calculate the configuration that will be used for this client when it is instantiated.
@@ -152,8 +183,13 @@ class Network:
     def aio_global(self) -> 'AIOGlobalClient':
         """
         Return a :class:`GlobalClient` that works on an asyncio event loop.
+
+        Note that once this object can only be accessed from the asyncio event loop it is intended
+        to be used on.
         """
-        return self._impl.global_impl(AIOGlobalClient)
+        client = self._impl.global_impl(AIOGlobalClient)
+        self._impl.freeze()
+        return client
 
     def simple_party(self, party: 'Union[str, Party]') -> 'SimplePartyClient':
         """
@@ -162,7 +198,14 @@ class Network:
 
         :param party: The party to get a client for.
         """
-        return self._impl.party_impl(Party(party), SimplePartyClient)
+        return self._impl.party_impl(to_party(party), SimplePartyClient)
+
+    def simple_new_party(self) -> 'SimplePartyClient':
+        """
+        Return a :class:`PartyClient` that exposes thread-safe, synchronous (blocking) methods for
+        communicating with a ledger. Callbacks are dispatched to background threads.
+        """
+        return self.simple_party(str(uuid4()))
 
     def aio_party(self, party: 'Union[str, Party]') -> 'AIOPartyClient':
         """
@@ -172,6 +215,13 @@ class Network:
         """
         return self._impl.party_impl(Party(party), AIOPartyClient)
 
+    def aio_new_party(self) -> 'AIOPartyClient':
+        """
+        Return a :class:`PartyClient` for a random party that works on an asyncio event loop.
+        This will never return the same object twice.
+        """
+        return self.aio_party(str(uuid4()))
+
     def party_bots(
             self,
             party: 'Union[str, Party]',
@@ -204,14 +254,19 @@ class Network:
             self._impl.invoker.install_signal_handlers()
         return self._impl.start(daemon)
 
-    def shutdown(self) -> None:
+    def shutdown(self) -> 'Optional[Awaitable[None]]':
         """
         Gracefully shut down all network connections and notify all clients that they are about to
         be terminated.
 
         The current thread does NOT block.
+
+        :return:
+            ``None`` unless ``start()`` was called, in which case the coroutine that
+            corresponds to dazl's "main" is returned.
         """
-        return self._impl.shutdown()
+        self._impl.shutdown()
+        return self._main_fut
 
     def join(self, timeout: 'Optional[float]' = None) -> None:
         """
@@ -226,6 +281,12 @@ class Network:
 
     # <editor-fold desc="asyncio-based scheduling API">
 
+    def start(self) -> None:
+        """
+        Start the coroutine that spawns callbacks for listeners on event streams.
+        """
+        self._main_fut = ensure_future(self.aio_run(keep_open=False))
+
     def run_until_complete(
             self, *coroutines: 'Awaitable[None]',
             install_signal_handlers: 'Optional[bool]' = None) \
@@ -293,6 +354,12 @@ class Network:
     def bots(self) -> 'Collection[Bot]':
         return self._impl.bots()
 
+    def __enter__(self):
+        pass
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.shutdown()
+
 
 class GlobalClient:
     """
@@ -310,7 +377,7 @@ class AIOGlobalClient(GlobalClient):
     async def ensure_dar(
             self,
             contents: 'Union[str, Path, bytes, BinaryIO]',
-            timeout: 'TimeDeltaConvertible' = DEFAULT_TIMEOUT_SECONDS) -> None:
+            timeout: 'TimeDeltaLike' = DEFAULT_TIMEOUT_SECONDS) -> None:
         """
         Validate that the ledger has the packages specified by the given contents (as a byte array).
         Throw an exception if the specified DARs do not exist within the specified timeout.
@@ -324,7 +391,7 @@ class AIOGlobalClient(GlobalClient):
     async def ensure_packages(
             self,
             package_ids: 'Collection[str]',
-            timeout: 'TimeDeltaConvertible' = DEFAULT_TIMEOUT_SECONDS) -> None:
+            timeout: 'TimeDeltaLike' = DEFAULT_TIMEOUT_SECONDS) -> None:
         """
         Validate that packages with the specified package IDs exist on the ledger. Throw an
         exception if the specified packages do not exist within the specified timeout.
@@ -340,19 +407,13 @@ class AIOGlobalClient(GlobalClient):
         """
         return await self._impl.aio_metadata()
 
-    async def get_time(self) -> datetime:
-        return await self._impl.get_time()
-
-    async def set_time(self, new_datetime: datetime) -> None:
-        await self._impl.set_time(new_datetime)
-
 
 class SimpleGlobalClient(GlobalClient):
 
     def ensure_dar(
             self,
             contents: 'Union[str, Path, bytes, BinaryIO]',
-            timeout: 'TimeDeltaConvertible' = DEFAULT_TIMEOUT_SECONDS) -> None:
+            timeout: 'TimeDeltaLike' = DEFAULT_TIMEOUT_SECONDS) -> None:
         """
         Validate that the ledger has the packages specified by the given contents (as a byte array).
         Throw an exception if the specified DARs do not exist within the specified timeout.
@@ -367,7 +428,7 @@ class SimpleGlobalClient(GlobalClient):
     def ensure_packages(
             self,
             package_ids: 'Collection[str]',
-            timeout: 'TimeDeltaConvertible' = DEFAULT_TIMEOUT_SECONDS) -> None:
+            timeout: 'TimeDeltaLike' = DEFAULT_TIMEOUT_SECONDS) -> None:
         """
         Validate that packages with the specified package IDs exist on the ledger. Throw an
         exception if the specified packages do not exist within the specified timeout.
@@ -378,19 +439,13 @@ class SimpleGlobalClient(GlobalClient):
         return self._impl.invoker.run_in_loop(
             lambda: self._impl.ensure_package_ids(package_ids, timeout))
 
-    def metadata(self, timeout: 'TimeDeltaConvertible' = DEFAULT_TIMEOUT_SECONDS) \
+    def metadata(self, timeout: 'TimeDeltaLike' = DEFAULT_TIMEOUT_SECONDS) \
             -> 'LedgerMetadata':
         """
         Return the current set of known packages.
         """
         return self._impl.simple_metadata(timeout)
 
-    def get_time(self) -> datetime:
-        return self._impl.invoker.run_in_loop(self._impl.get_time)
-
-    def set_time(self, new_datetime: datetime) -> None:
-        self._impl.invoker.run_in_loop(lambda: self._impl.set_time(new_datetime))
-
 
 class PartyClient:
     """
@@ -657,7 +712,7 @@ class AIOPartyClient(PartyClient):
             self,
             commands: 'EventHandlerResponse',
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> 'Awaitable[None]':
         """
         Submit commands to the ledger.
@@ -666,20 +721,23 @@ class AIOPartyClient(PartyClient):
             An object that can be converted to a command.
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         :return:
             A future that resolves when the command has made it to the ledger _or_ an error
             occurred when trying to process them.
         """
-        return self._impl.write_commands(commands, workflow_id=workflow_id, ttl=ttl)
+        return self._impl.write_commands(
+            commands, workflow_id=workflow_id, deduplication_time=deduplication_time)
 
     def submit_create(
             self,
             template_name: str,
             arguments: 'Optional[dict]' = None,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> 'Awaitable[None]':
         """
         Submit a single create command. Equivalent to calling :meth:`submit` with a single
@@ -691,14 +749,19 @@ class AIOPartyClient(PartyClient):
             The arguments to the create (as a ``dict``).
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         :return:
             A future that resolves when the command has made it to the ledger _or_ an error
             occurred when trying to process them.
         """
         from .. import create
-        return self.submit(create(template_name, arguments), workflow_id=workflow_id, ttl=ttl)
+        return self.submit(
+            create(template_name, arguments),
+            workflow_id=workflow_id,
+            deduplication_time=deduplication_time)
 
     def submit_exercise(
             self,
@@ -706,7 +769,7 @@ class AIOPartyClient(PartyClient):
             choice_name: str,
             arguments: 'Optional[dict]' = None,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> 'Awaitable[None]':
         """
         Submit a single exercise choice. Equivalent to calling :meth:`submit` with a single
@@ -721,14 +784,19 @@ class AIOPartyClient(PartyClient):
             choices.
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         :return:
             A future that resolves when the command has made it to the ledger _or_ an error
             occurred when trying to process them.
         """
         from .. import exercise
-        return self.submit(exercise(cid, choice_name, arguments), workflow_id=workflow_id, ttl=ttl)
+        return self.submit(
+            exercise(cid, choice_name, arguments),
+            workflow_id=workflow_id,
+            deduplication_time=deduplication_time)
 
     def submit_exercise_by_key(
             self,
@@ -737,7 +805,7 @@ class AIOPartyClient(PartyClient):
             choice_name: str,
             arguments: 'Optional[dict]' = None,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> 'Awaitable[None]':
         """
         Synchronously submit a single exercise choice. Equivalent to calling :meth:`submit` with a
@@ -754,14 +822,16 @@ class AIOPartyClient(PartyClient):
             choices.
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         """
         from .. import exercise_by_key
         return self.submit(
             exercise_by_key(template_name, contract_key, choice_name, arguments),
             workflow_id=workflow_id,
-            ttl=ttl)
+            deduplication_time=deduplication_time)
 
     def submit_create_and_exercise(
             self,
@@ -770,7 +840,7 @@ class AIOPartyClient(PartyClient):
             choice_name: str,
             choice_arguments: 'Optional[dict]' = None,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> 'Awaitable[None]':
         """
         Synchronously submit a single create-and-exercise command. Equivalent to calling
@@ -786,13 +856,16 @@ class AIOPartyClient(PartyClient):
             The arguments to the exercise (as a ``dict``). Can be omitted (``None``) for no-argument
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         """
         from .. import create_and_exercise
         return self.submit(
             create_and_exercise(template_name, arguments, choice_name, choice_arguments),
-            workflow_id=workflow_id)
+            workflow_id=workflow_id,
+            deduplication_time=deduplication_time)
 
     # </editor-fold>
 
@@ -901,19 +974,19 @@ class AIOPartyClient(PartyClient):
     def set_config(self, url: 'Optional[str]', **kwargs):
         self._impl.set_config(url=url, **kwargs)
 
-    def get_time(self) -> 'Awaitable[datetime]':
+    async def ensure_dar(
+            self,
+            contents: 'Union[str, Path, bytes, BinaryIO]',
+            timeout: 'TimeDeltaLike' = DEFAULT_TIMEOUT_SECONDS) -> None:
         """
-        Return the current time on the remote server. Also advance the local notion of time if
-        required.
-        """
-        return self._impl.get_time()
+        Validate that the ledger has the packages specified by the given contents (as a byte array).
+        Throw an exception if the specified DARs do not exist within the specified timeout.
 
-    def set_time(self, new_datetime: datetime) -> 'Awaitable[None]':
+        :param contents: The DAR or DALF to ensure.
+        :param timeout: The maximum length of time to wait before giving up.
         """
-        Set the current time on the ledger. This is only supported if the ledger supports time
-        manipulation.
-        """
-        return self._impl.set_time(new_datetime)
+        raw_bytes = get_bytes(contents)
+        return await self._impl.parent.upload_package(raw_bytes, timeout)
 
     def ready(self) -> 'Awaitable[None]':
         """
@@ -1213,7 +1286,7 @@ class SimplePartyClient(PartyClient):
             self,
             commands,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> None:
         """
         Submit commands to the ledger.
@@ -1222,21 +1295,24 @@ class SimplePartyClient(PartyClient):
             An object that can be converted to a command.
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         :return:
             A future that resolves when the command has made it to the ledger _or_ an error
             occurred when trying to process them.
         """
         return self._impl.invoker.run_in_loop(
-            lambda: self._impl.write_commands(commands, workflow_id=workflow_id, ttl=ttl))
+            lambda: self._impl.write_commands(
+                commands, workflow_id=workflow_id, deduplication_time=deduplication_time))
 
     def submit_create(
             self,
             template_name: 'TemplateNameLike',
             arguments: 'Optional[dict]' = None,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> None:
         """
         Synchronously submit a single create command. Equivalent to calling :meth:`submit` with a
@@ -1248,11 +1324,16 @@ class SimplePartyClient(PartyClient):
             The arguments to the create (as a ``dict``).
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         """
         from .. import create
-        return self.submit(create(template_name, arguments), workflow_id=workflow_id, ttl=ttl)
+        return self.submit(
+            create(template_name, arguments),
+            workflow_id=workflow_id,
+            deduplication_time=deduplication_time)
 
     def submit_exercise(
             self,
@@ -1260,7 +1341,7 @@ class SimplePartyClient(PartyClient):
             choice_name: str,
             arguments: 'Optional[dict]' = None,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> None:
         """
         Synchronously submit a single exercise choice. Equivalent to calling :meth:`submit` with a
@@ -1275,11 +1356,16 @@ class SimplePartyClient(PartyClient):
             choices.
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         """
         from .. import exercise
-        return self.submit(exercise(cid, choice_name, arguments), workflow_id=workflow_id, ttl=ttl)
+        return self.submit(
+            exercise(cid, choice_name, arguments),
+            workflow_id=workflow_id,
+            deduplication_time=deduplication_time)
 
     def submit_exercise_by_key(
             self,
@@ -1288,7 +1374,7 @@ class SimplePartyClient(PartyClient):
             choice_name: str,
             arguments: 'Optional[dict]' = None,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> None:
         """
         Synchronously submit a single exercise choice. Equivalent to calling :meth:`submit` with a
@@ -1303,15 +1389,18 @@ class SimplePartyClient(PartyClient):
         :param arguments:
             The arguments to the create (as a ``dict``). Can be omitted (``None``) for no-argument
             choices.
-        :param ttl:
-            A time-to-live for the command.
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         """
         from .. import exercise_by_key
         return self.submit(
             exercise_by_key(template_name, contract_key, choice_name, arguments),
-            workflow_id=workflow_id, ttl=ttl)
+            workflow_id=workflow_id,
+            deduplication_time=deduplication_time)
 
     def submit_create_and_exercise(
             self,
@@ -1320,7 +1409,7 @@ class SimplePartyClient(PartyClient):
             choice_name: str,
             choice_arguments: 'Optional[dict]' = None,
             workflow_id: 'Optional[str]' = None,
-            ttl: 'Optional[TimeDeltaConvertible]' = None) \
+            deduplication_time: 'Optional[TimeDeltaLike]' = None) \
             -> None:
         """
         Synchronously submit a single create-and-exercise command. Equivalent to calling
@@ -1336,13 +1425,16 @@ class SimplePartyClient(PartyClient):
             The arguments to the exercise (as a ``dict``). Can be omitted (``None``) for no-argument
         :param workflow_id:
             The optional workflow ID to stamp on the outgoing command.
-        :param ttl:
-            A time-to-live for the command.
+        :param deduplication_time:
+            The length of the time window during which all commands with the same party and command
+            ID will be deduplicated. Duplicate commands submitted before the end of this window
+            return an ``ALREADY_EXISTS`` error.
         """
         from .. import create_and_exercise
         return self.submit(
             create_and_exercise(template_name, arguments, choice_name, choice_arguments),
-            workflow_id=workflow_id)
+            workflow_id=workflow_id,
+            deduplication_time=deduplication_time)
 
     # endregion
 
@@ -1456,11 +1548,20 @@ class SimplePartyClient(PartyClient):
     def set_config(self, url: Optional[str], **kwargs):
         self._impl.set_config(url=url, **kwargs)
 
-    def get_time(self) -> datetime:
-        return self._impl.invoker.run_in_loop(lambda: self._impl.get_time())
+    def ensure_dar(
+            self,
+            contents: 'Union[str, Path, bytes, BinaryIO]',
+            timeout: 'TimeDeltaLike' = DEFAULT_TIMEOUT_SECONDS) -> None:
+        """
+        Validate that the ledger has the packages specified by the given contents (as a byte array).
+        Throw an exception if the specified DARs do not exist within the specified timeout.
 
-    def set_time(self, new_datetime: datetime) -> None:
-        return self._impl.invoker.run_in_loop(lambda: self._impl.set_time(new_datetime))
+        :param contents: The DAR or DALF to ensure.
+        :param timeout: The maximum length of time to wait before giving up.
+        """
+        raw_bytes = get_bytes(contents)
+        return self._impl.invoker.run_in_loop(
+            lambda: self._impl.parent.upload_package(raw_bytes, timeout))
 
     def ready(self) -> None:
         """
