@@ -1,14 +1,17 @@
-# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
+# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
 # SPDX-License-Identifier: Apache-2.0
 
 
+from typing import Optional, Sequence
+
 from .daml_lf_1 import *
-from ..model.types import ModuleRef, TypeReference
+
+__all__ = ['ProtobufParser']
 
 
 # noinspection PyPep8Naming,PyMethodMayBeStatic
 class ProtobufParser:
-    def __init__(self, current_package: str):
+    def __init__(self, current_package: 'PackageRef'):
         from typing import List
         self.current_package = current_package
         self.interned_strings = []  # type: List[str]
@@ -24,17 +27,22 @@ class ProtobufParser:
         if sum_name is None:
             return None
         elif sum_name == 'self':
-            return ModuleRef(self.current_package, module_name.segments)
+            return ModuleRef(self.current_package, module_name)
         elif sum_name == 'package_id_str':
-            return ModuleRef(pb.package_ref.package_id_str, module_name.segments)
+            return ModuleRef(pb.package_ref.package_id_str, module_name)
         elif sum_name == 'package_id_interned_str':
             return ModuleRef(
-                self.interned_strings[pb.package_ref.package_id_interned_str], module_name.segments)
+                self.interned_strings[pb.package_ref.package_id_interned_str], module_name)
         else:
             raise ValueError(f'unknown sum type value: {sum_name!r}')
 
-    def parse_TypeConName(self, pb) -> 'TypeReference':
-        return TypeReference(
+    def parse_TypeConName(self, pb) -> 'TypeConName':
+        return TypeConName(
+            self.parse_ModuleRef(pb.module),
+            self._resolve_dotted_name(pb.name_dname, pb.name_interned_dname).segments)
+
+    def parse_TypeSynName(self, pb) -> 'TypeSynName':
+        return TypeSynName(
             self.parse_ModuleRef(pb.module),
             self._resolve_dotted_name(pb.name_dname, pb.name_interned_dname).segments)
 
@@ -110,6 +118,8 @@ class ProtobufParser:
             return Type(tuple=self.parse_Type_Tuple(pb.struct))
         elif sum_name == 'nat':
             return Type(nat=pb.nat)
+        elif sum_name == 'syn':
+            return Type(syn=self.parse_Type_Syn(pb.syn))
         else:
             raise ValueError(f'unknown sum type value: {sum_name!r}')
 
@@ -148,6 +158,11 @@ class ProtobufParser:
     def parse_Type_Tuple(self, pb) -> 'Type.Tuple':
         return Type.Tuple(tuple(self.parse_FieldWithType(field) for field in pb.fields))
 
+    def parse_Type_Syn(self, pb) -> 'Type.Syn':
+        return Type.Syn(
+            tysyn=self.parse_TypeSynName(pb.tysyn),
+            args=tuple(self.parse_Type(arg) for arg in pb.args))
+
     def parse_PrimCon(self, pb) -> 'Optional[PrimCon]':
         if pb is None:
             return None
@@ -451,7 +466,6 @@ class ProtobufParser:
             template=self.parse_TypeConName(pb.template),
             choice=self._resolve_string(pb.choice_str, pb.choice_interned_str),
             cid=self.parse_Expr(pb.cid),
-            actor=self.parse_Expr(pb.actor, optional=True),
             arg=self.parse_Expr(pb.arg))
 
     def parse_Update_Fetch(self, pb) -> 'Update.Fetch':
@@ -601,6 +615,13 @@ class ProtobufParser:
         ctors_2 = tuple(self.interned_strings[idx] for idx in pb.constructors_interned_str)
         return DefDataType.EnumConstructors(constructors=ctors_1 + ctors_2)
 
+    def parse_DefTypeSyn(self, pb) -> 'DefTypeSyn':
+        return DefTypeSyn(
+            name=self._resolve_dotted_name(pb.name_dname, pb.name_interned_dname),
+            params=tuple(self.parse_TypeVarWithKind(param) for param in pb.params),
+            type=self.parse_Type(pb.type),
+            location=self.parse_Location(pb.location))
+
     def parse_DefValue(self, pb) -> 'DefValue':
         return DefValue(
             name_with_type=self.parse_DefValue_NameWithType(pb.name_with_type),
@@ -624,6 +645,7 @@ class ProtobufParser:
         return Module(
             name=self._resolve_dotted_name(pb.name_dname, pb.name_interned_dname),
             flags=self.parse_FeatureFlags(pb.flags),
+            synonyms=tuple(self.parse_DefTypeSyn(value) for value in pb.synonyms),
             data_types=tuple(self.parse_DefDataType(data_type) for data_type in pb.data_types),
             values=tuple(self.parse_DefValue(value) for value in pb.values),
             templates=tuple(self.parse_DefTemplate(template) for template in pb.templates))
@@ -639,7 +661,14 @@ class ProtobufParser:
 
         self.interned_dotted_names.extend(indices)
 
-        return Package(modules=tuple(self.parse_Module(module) for module in pb.modules))
+        return Package(
+            modules=tuple(self.parse_Module(module) for module in pb.modules),
+            metadata=self.parse_PackageMetadata(pb.metadata) if pb.HasField('metadata') else None)
+
+    def parse_PackageMetadata(self, pb) -> 'PackageMetadata':
+        return PackageMetadata(
+            name=self.interned_strings[pb.name_interned_str],
+            version=self.interned_strings[pb.version_interned_str])
 
     def _resolve_string(self, name: 'Optional[str]', interned_id: 'Optional[int]') -> str:
         return name if name else self.interned_strings[interned_id]
