@@ -1,4 +1,4 @@
-# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
+# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
 # SPDX-License-Identifier: Apache-2.0
 
 """
@@ -17,7 +17,7 @@ from typing import Any, Awaitable, Callable, Generator, Generic, Iterable, List,
     Sequence, TypeVar, Union
 
 from .. import LOG
-from .prim_types import to_timedelta, TimeDeltaConvertible
+from ..prim.datetime import TimeDeltaLike, to_timedelta
 
 T = TypeVar('T', covariant=True)
 U = TypeVar('U')
@@ -27,7 +27,6 @@ _CANCELLED = 'CANCELLED'
 _FINISHED = 'FINISHED'
 
 
-
 def non_reentrant(async_fn):
     if not iscoroutinefunction(async_fn):
         raise ValueError('expected a coroutine function')
@@ -163,7 +162,7 @@ class FailedInvocation:
 def execute_in_loop(
         loop: AbstractEventLoop,
         coro_fn: 'Callable[[], Union[Awaitable[T], T]]',
-        timeout: 'Optional[TimeDeltaConvertible]' = 30.0) -> T:
+        timeout: 'Optional[TimeDeltaLike]' = 30.0) -> T:
     """
     Run a coroutine in a target loop. Exceptions thrown by the coroutine are
     propagated to the caller. Must NOT be called from a coroutine on the same
@@ -537,7 +536,9 @@ class ContextFreeFuture(Awaitable[T]):
             self._schedule_callbacks()
             self._log_traceback = True
 
-    def add_done_callback(self, fn: 'Callable[[ContextFreeFuture[T]], None]') -> None:
+    def add_done_callback(self, fn: 'Callable[[ContextFreeFuture[T]], None]', context=None) -> None:
+        if context is not None:
+            LOG.warning("ContextFreeFutures do not support the use of contexts.")
         with self._lock:
             if self._loop is not None and self._state != _PENDING:
                 self._loop.call_soon_threadsafe(fn, self)
@@ -663,8 +664,11 @@ class DeferredStartTask:
     def exception(self) -> Optional[BaseException]:
         return self._future.exception()
 
-    def add_done_callback(self, callback: Callable[[Future], Any]) -> None:
-        self._future.add_done_callback(callback)
+    def add_done_callback(self, callback: 'Callable[[Future], Any]', context=None) -> None:
+        if context is None:
+            self._future.add_done_callback(callback)
+        else:
+            self._future.add_done_callback(callback, context=context)
 
     def __await__(self):
         return self._future.__await__()
